<?php
/**
 * dtweb - Device Tree editor for UDOO boards
 * Copyright (C) 2015 Francesco Montefoschi <francesco.monte@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package dtweb
 * @author  Francesco Montefoschi
 * @license http://www.gnu.org/licenses/gpl-3.0.html  GNU GPL 3.0
 */

class Controller_Save extends Controller
{
    public function run()
    {
        $origDtsDir = realpath(__DIR__ . '/../../dts');
        $url = parse_url($_SERVER['HTTP_REFERER']);
        parse_str($url['query'], $params);
        $kernelDir = $params['kerneldir'];
        $board = $params['board'];
        $mkswu = 0;

        switch ($board) {
            case 'Armadillo-640':
                $orig = "$origDtsDir/armadillo-640-expansion-interface.dtsi";
                $dtsi = "$kernelDir/arch/arm/boot/dts/armadillo-640-expansion-interface.dtsi";
                $dts_dir = "$kernelDir/arch/arm/boot/dts";
                $target = "armadillo-640-at-dtweb.dtb";
                $cores = 1;
                $flags = "ARCH=arm CROSS_COMPILE=arm-linux-gnueabihf-";
                break;
            case 'Armadillo-610':
                $orig = "$origDtsDir/armadillo-610-expansion-interface.dtsi";
                $dtsi = "$kernelDir/arch/arm/boot/dts/armadillo-610-expansion-interface.dtsi";
                $dts_dir = "$kernelDir/arch/arm/boot/dts";
                $target = "armadillo-610-at-dtweb.dtb";
                $cores = 1;
                $flags = "ARCH=arm CROSS_COMPILE=arm-linux-gnueabihf-";
                break;
            case 'Armadillo-IoT_A6':
                $orig = "$origDtsDir/armadillo-iotg-a6-expansion-interface.dtsi";
                $dtsi = "$kernelDir/arch/arm/boot/dts/armadillo-iotg-a6-expansion-interface.dtsi";
                $dts_dir = "$kernelDir/arch/arm/boot/dts";
                $target = "armadillo-iotg-a6-at-dtweb.dtb";
                $cores = 1;
                $flags = "ARCH=arm CROSS_COMPILE=arm-linux-gnueabihf-";
                break;
            case 'Armadillo-IoT_G4':
                $orig = "$origDtsDir/armadillo_iotg_g4-expansion-interface.dtsi";
                $dtsi = "$kernelDir/arch/arm64/boot/dts/freescale/armadillo_iotg_g4-expansion-interface.dtsi";
                $dts_dir = "$kernelDir/arch/arm64/boot/dts";
                $target = "freescale/armadillo_iotg_g4-at-dtweb.dtbo";
                $cores = 1;
                $flags = "ARCH=arm64 CROSS_COMPILE=aarch64-linux-gnu-";
                $mkswu = 1;
                break;
            default:
                $this->json(array('success' => false, 'message' => "Invalid board model: " . $_POST['id']));
        }
        
        $this->build($orig, $dtsi, $dts_dir, $target, $cores, $flags, $mkswu);
    }
    
    private function fail($message)
    {
        $this->json(array(
            'success' => false,
            'message' => "Error: " . $message
        ));
    }

    private function build($orig, $dtsi, $dts_dir, $target, $cores, $flags, $mkswu)
    {
        $kernelDir = Service_KernelDetector::kernelFromURL();

        $configDir = $kernelDir;
        
        $dtreader = new Service_DeviceTreeReader();
        $dt = $dtreader->parse($orig);

        $conf = json_decode($_POST['conf'], true);
        $prop = json_decode($_POST['prop'], true);
        $snvspins = array();
        if (isset($_POST['snvspins'])) {
            $snvspins = json_decode($_POST['snvspins'], true);
        }
        $config = new Model_Configuration();
        foreach ($conf as $name => $pins) {
            $config->add($name, $pins);
        }

        $dteditor = new Service_DeviceTreeEditor($dt);
        $dteditor->setBoardType($_POST['id']);
        $dteditor->disableEverything();
        $dteditor->applyConfiguration($config, $prop, $snvspins);

        file_put_contents($dtsi, $dteditor->generate());

        /* remove dummy node */
        exec("bash -c 'sed -i \"/DUMMY_NODES_START/,/DUMMY_NODES_END/d\" $dtsi'");
        /* remove comment line */
        exec("bash -c 'sed -i \"/^\/\//d\" $dtsi'");
        /* combine multiple newlines */
        exec("bash -c 'sed -i \"/^$/{N; /^\\n$/D;}\" $dtsi'");
        /* remove simbol of dtweb */
        exec("bash -c 'sed -i \"s/{{.\+}}//g\" $dtsi'");
        /* remove discarded comment */
        exec("bash -c 'sed -i \"s/\/\/ $//g\" $dtsi'");
        /* remove trailing whitespace */
        exec("bash -c 'sed -i \"s/[ \\t]*$//g\" $dtsi'");

        chdir(dirname($dtsi));
        exec("bash -c 'rm -f $target'");
        
        chdir($kernelDir);
        exec("bash -c 'make $flags -j$cores $target 2>&1'", $output, $returnCode);

        if ($returnCode != 0) {
            $this->fail('Cannot build DTB! Error log: <br><br><code>' . implode('<br>', $output) . '</code>');
            return;
        }

        file_put_contents("$configDir/.dtweb." . $_POST['id'] . ".conf", $_POST['conf']);
        file_put_contents("$configDir/.dtweb.prop." . $_POST['id'] . ".conf", $_POST['prop']);

        if ($mkswu) {
            if (substr($target, -1) != "o") {
                $this->fail("DTB built, but can only prepare mkswu config for overlays");
                return;
            }
            /* get home from kerneldir */
            $home = implode("/", array_slice(explode("/", $kernelDir), 0, 3));
            $mkswu_dir = "$home/mkswu/at-dtweb";
            /* get the dtb filename from build target */
            $dtb = explode("/", $target);
            $dtb = end($dtb);
            exec("
                 set -e
                 mkdir -p \"$mkswu_dir\"
                 # copy created dtb
                 cp \"$dts_dir/$target\" \"$mkswu_dir/\"

                 # link to update scripts in example dir
                 # this allows doing this now even if mkswu is not installed yet
                 ln -fs /usr/share/mkswu/examples/update_preserve_files.sh \"$mkswu_dir/\"
                 ln -fs /usr/share/mkswu/examples/update_overlays.sh \"$mkswu_dir/\"

                 # Do not overwrite if previously generated
                 version=1
                 if [ -e \"$mkswu_dir/at-dtweb.desc\" ]; then
                     version=\$(awk -F= '$1 == \"version\" { print $2 + 1 }' \
                                \"$mkswu_dir/at-dtweb.desc\")
                     mv \"$mkswu_dir/at-dtweb.desc\" \"$mkswu_dir/at-dtweb.desc.old\"
                 fi
                 cat > \"$mkswu_dir/at-dtweb.desc\" <<EODESC
component=extra_os.at_dtweb
version=\${version:-1}

swdesc_files --dest /boot \"$dtb\"
swdesc_script update_overlays.sh -- \"$dtb\"
swdesc_script update_preserve_files.sh -- \"/boot/$dtb\"
EODESC
                 ", $output, $returnCode);
            if ($returnCode != 0) {
                $this->fail("DTB overlay built at $dts_dir/$target, but could not create mkswu .desc configuration:<br><br><code>" . implode('<br>', $output) . '</code>');
                return;
            }

            $this->json(array(
                'success' => true,
                'message' => "Device tree overlay copied to $mkswu_dir/$dtb, SWU description file $mkswu_dir/at-dtweb.desc ready to be used"
            ));
        } else {
            $this->json(array(
                'success' => true,
                'message' => "Device tree built at $dts_dir/$target"
            ));
        }
    }
}
